/*
 * Copyright 2014 Trend Micro Incorporated
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/
 */

#include <linux/kernel.h>
#include "tc_common.h"

struct u32_sel
{
	struct tc_u32_sel sel;
	struct tc_u32_key keys[128];
};

static int get_u32_handle(__u32 *handle, const char *str)
{
	__u32 htid=0, hash=0, nodeid=0;
	char *tmp = strchr(str, ':');

	if (tmp == NULL) {
		if (memcmp("0x", str, 2) == 0)
			return get_u32(handle, str, 16);
		return -1;
	}
	htid = strtoul(str, &tmp, 16);
	if (tmp == str && *str != ':' && *str != 0)
		return -1;
	if (htid>=0x1000)
		return -1;
	if (*tmp) {
		str = tmp+1;
		hash = strtoul(str, &tmp, 16);
		if (tmp == str && *str != ':' && *str != 0)
			return -1;
		if (hash>=0x100)
			return -1;
		if (*tmp) {
			str = tmp+1;
			nodeid = strtoul(str, &tmp, 16);
			if (tmp == str && *str != 0)
				return -1;
			if (nodeid>=0x1000)
				return -1;
		}
	}
	*handle = (htid<<20)|(hash<<12)|nodeid;
	return 0;
}

static int parse_mark(int *argc_p, char ***argv_p, struct nlmsghdr *n)
{
	int res = -1;
	int argc = *argc_p;
	char **argv = *argv_p;
	struct tc_u32_mark mark;

	if (argc <= 1)
		return -1;

	if (get_u32(&mark.val, *argv, 0)) {
		TC_ERR("Illegal \"mark\" value");
		return -1;
	}
	TC_DBG("get mark value 0x%08x", mark.val);
	NEXT_ARG();

	if (get_u32(&mark.mask, *argv, 0)) {
		TC_ERR("Illegal \"mark\" mask");
		return -1;
	}
	TC_DBG("get mark mask 0x%08x", mark.mask);
	NEXT_ARG();

	if ((mark.val & mark.mask) != mark.val) {
		TC_ERR("Illegal \"mark\" (impossible combination)");
		return -1;
	}

	addattr_l(n, MAX_MSG, TCA_U32_MARK, &mark, sizeof(mark));
	res = 0;

	*argc_p = argc;
	*argv_p = argv;
	return res;
}

static int parse_selector(int *argc_p, char ***argv_p,
			  struct tc_u32_sel *sel, struct nlmsghdr *n)
{
	int argc = *argc_p;
	char **argv = *argv_p;
	int res = -1;

	if (argc <= 0)
		return -1;

	if (strcmp(*argv, "mark") == 0)
	{
		NEXT_ARG();
		res = parse_mark(&argc, &argv, n);
	}
	else
		return -1;

	*argc_p = argc;
	*argv_p = argv;
	return res;
}


int u32_parse_filter_opt(struct filter_util *fu, char *handle, int argc, char **argv, struct nlmsghdr *n)
{
	struct u32_sel *sel;
	struct tcmsg *t = NLMSG_DATA(n);
	struct rtattr *tail;

	int sel_ok = 0;

	alloc_mem(sel, sizeof(*sel), GFP_KERNEL);
	memset(sel, 0, sizeof(*sel));

	if (handle && get_u32_handle(&t->tcm_handle, handle))
	{
		TC_ERR("Illegal filter ID");
		goto error_exit;
	}

	if (argc <= 0)
		goto exit;

	tail = NLMSG_TAIL(n);
	addattr_l(n, MAX_MSG, TCA_OPTIONS, NULL, 0);

	while (argc > 0)
	{
		if (strcmp(*argv, "match") == 0)
		{
			NEXT_ARG();
			if (parse_selector(&argc, &argv, &sel->sel, n))
			{
				TC_ERR("Illegal \"match\"");
				goto error_exit;
			}
			sel_ok++;
			continue;
		}
		else if (strcmp(*argv, "classid") == 0 ||
			strcmp(*argv, "flowid") == 0)
		{
			unsigned h;
			NEXT_ARG();
			if (get_handle(&h, *argv))
			{
				TC_ERR("Illegal \"classid\"");
				goto error_exit;
			}
			addattr_l(n, MAX_MSG, TCA_U32_CLASSID, &h, 4);
			sel->sel.flags |= TC_U32_TERMINAL;
		}
		else
		{
			TC_ERR("What is \"%s\"?", *argv);
			goto error_exit;
		}
		argc--; argv++;
	}

	if (sel_ok)
	{
		addattr_l(n, MAX_MSG, TCA_U32_SEL, sel,
			sizeof(sel->sel) + sel->sel.nkeys * sizeof(struct tc_u32_key));
	}

	tail->rta_len = (void *) NLMSG_TAIL(n) - (void *) tail;

exit:
	free_mem(sel);
	return 0;

error_exit:
	free_mem(sel);
	return -1;
}
