/*
 * Copyright 2014 Trend Micro Incorporated
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/
 */

#include <linux/kernel.h>
#include "tc_common.h"

#define explain1(_str) TC_ERR("Invalid " _str " \"%s\".", *argv);

int htb_parse_opt(struct qdisc_util *qu, int argc, char **argv, struct nlmsghdr *n)
{
	struct tc_htb_glob opt;
	struct rtattr *tail;

	memset(&opt, 0, sizeof(opt));
	opt.rate2quantum = 10;
	opt.version = 3;

	while (argc > 0)
	{
		if (matches(*argv, "r2q") == 0)
		{
			NEXT_ARG();
			if (get_u32(&opt.rate2quantum, *argv, 10))
			{
				TC_ERR("Invalid r2q value \"%s\".", *argv);
				return -1;
			}
		}
		else if (matches(*argv, "default") == 0)
		{
			NEXT_ARG();
			if (get_u32(&opt.defcls, *argv, 16))
			{
				TC_ERR("Invalid default class \"%s\".", *argv);
				return -1;
			}
		}
		else
		{
			TC_ERR("What is \"%s\"?", *argv);
			return -1;
		}
		argc--;
		argv++;
	}

	tail = NLMSG_TAIL(n);
	addattr_l(n, 1024, TCA_OPTIONS, NULL, 0);
	addattr_l(n, 2024, TCA_HTB_INIT, &opt, NLMSG_ALIGN(sizeof(opt)));
	tail->rta_len = (void *) NLMSG_TAIL(n) - (void *) tail;
	return 0;
}

int htb_parse_class_opt(struct qdisc_util *qu, int argc, char **argv, struct nlmsghdr *n)
{
	int ok=0;
	struct __tc_htb_opt opt;
	u32 *rtab;
	u32 *ctab;
	unsigned buffer=0,cbuffer=0;
	int cell_log=-1,ccell_log = -1;
	unsigned mtu;
	unsigned short mpu = 0;
	unsigned short overhead = 0;
//	unsigned int direct_qlen = ~0U;
	unsigned int linklayer  = LINKLAYER_ETHERNET; /* Assume ethernet */
	struct rtattr *tail;
	u32 ceil = 0, rate = 0;

	alloc_mem(rtab, 256 * sizeof(*rtab), GFP_KERNEL);
	alloc_mem(ctab, 256 * sizeof(*ctab), GFP_KERNEL);

	memset(&opt, 0, sizeof(opt));
	mtu = 1600; /* eth packet len */

	while (argc > 0) {
		if (matches(*argv, "prio") == 0) {
			NEXT_ARG();
			if (get_u32(&opt.prio, *argv, 10)) {
				explain1("prio");
				goto error_exit;
			}
			ok++;
		} else if (matches(*argv, "mtu") == 0) {
			NEXT_ARG();
			if (get_u32(&mtu, *argv, 10)) {
				explain1("mtu");
				goto error_exit;
			}
#if 0
		} else if (matches(*argv, "direct_qlen") == 0) {
			NEXT_ARG();
			if (get_u32(&direct_qlen, *argv, 10)) {
				explain1("direct_qlen");
				goto error_exit;
			}
#endif
		} else if (matches(*argv, "mpu") == 0) {
			NEXT_ARG();
			if (get_u16(&mpu, *argv, 10)) {
				explain1("mpu");
				goto error_exit;
			}
		} else if (matches(*argv, "overhead") == 0) {
			NEXT_ARG();
			if (get_u16(&overhead, *argv, 10)) {
				explain1("overhead");
				goto error_exit;
			}
#if 0
		} else if (matches(*argv, "linklayer") == 0) {
			NEXT_ARG();
			if (get_linklayer(&linklayer, *argv)) {
				explain1("linklayer"); return -1;
			}
#endif
		} else if (matches(*argv, "quantum") == 0) {
			NEXT_ARG();
			if (get_u32(&opt.quantum, *argv, 10)) {
				explain1("quantum");
				goto error_exit;
			}
		} else if (matches(*argv, "burst") == 0 ||
			   strcmp(*argv, "buffer") == 0 ||
			   strcmp(*argv, "maxburst") == 0) {
			NEXT_ARG();
			if (get_size_and_cell(&buffer, &cell_log, *argv) < 0) {
				explain1("buffer");
				goto error_exit;
			}
			ok++;
		} else if (matches(*argv, "cburst") == 0 ||
			   strcmp(*argv, "cbuffer") == 0 ||
			   strcmp(*argv, "cmaxburst") == 0) {
			NEXT_ARG();
			if (get_size_and_cell(&cbuffer, &ccell_log, *argv) < 0) {
				explain1("cbuffer");
				goto error_exit;
			}
			ok++;
		} else if (strcmp(*argv, "ceil") == 0) {
			NEXT_ARG();
			if (ceil) {
				TC_ERR("Double \"ceil\" spec.");
				goto error_exit;
			}
			if (get_rate(&ceil, *argv)) {
				explain1("ceil");
				goto error_exit;
			}
			ok++;
		} else if (strcmp(*argv, "rate") == 0) {
			NEXT_ARG();
			if (rate) {
				TC_ERR("Double \"rate\" spec.");
				goto error_exit;
			}
			if (get_rate(&rate, *argv)) {
				explain1("rate");
				goto error_exit;
			}
			ok++;
#if 0
		} else if (strcmp(*argv, "help") == 0) {
			explain();
			return -1;
#endif
		} else {
			TC_ERR("What is \"%s\"?", *argv);
			goto error_exit;
		}
		argc--; argv++;
	}

	/*	if (!ok)
		return 0;*/

	if (!rate) {
		TC_ERR("\"rate\" is required.");
		goto error_exit;
	}
	/* if ceil params are missing, use the same as rate */
	if (!ceil)
		ceil = rate;

	opt.rate.rate = (rate >= (1ULL << 32)) ? ~0U : rate;
	opt.ceil.rate = (ceil >= (1ULL << 32)) ? ~0U : ceil;

	/* compute minimal allowed burst from rate; mtu is added here to make
	   sute that buffer is larger than mtu and to have some safeguard space */
	if (!buffer)
		buffer = rate / get_hz() + mtu;
	if (!cbuffer)
		cbuffer = ceil / get_hz() + mtu;

	TC_DBG("rate %u hz %u mtu %u => buffer %u", opt.rate.rate, get_hz(), mtu, buffer);
	TC_DBG("ceil %u hz %u mtu %u => cbuffer %u", opt.ceil.rate, get_hz(), mtu, cbuffer);

	opt.ceil.overhead = overhead;
	opt.rate.overhead = overhead;

	opt.ceil.mpu = mpu;
	opt.rate.mpu = mpu;

	if (tc_calc_rtable(&opt.rate, rtab, cell_log, mtu, linklayer) < 0) {
		TC_ERR("htb: failed to calculate rate table.");
		goto error_exit;
	}
	opt.buffer = tc_calc_xmittime(rate, buffer);
	//tc_calc_xmittime(&opt.buffer, rate, buffer);
	TC_DBG("burst = %u", opt.buffer);

	if (tc_calc_rtable(&opt.ceil, ctab, ccell_log, mtu, linklayer) < 0) {
		TC_ERR("htb: failed to calculate ceil rate table.");
		goto error_exit;
	}
	opt.cbuffer = tc_calc_xmittime(ceil, cbuffer);
	//tc_calc_xmittime(&opt.cbuffer, ceil, cbuffer);
	TC_DBG("cburst = %u", opt.cbuffer);

	tail = NLMSG_TAIL(n);
#if 0
	if (direct_qlen != ~0U)
		addattr_l(n, 1024, TCA_HTB_DIRECT_QLEN,
			  &direct_qlen, sizeof(direct_qlen));
#endif
	addattr_l(n, 1024, TCA_OPTIONS, NULL, 0);

#if 0
	if (rate64 >= (1ULL << 32))
		addattr_l(n, 1124, TCA_HTB_RATE64, &rate64, sizeof(rate64));

	if (ceil64 >= (1ULL << 32))
		addattr_l(n, 1224, TCA_HTB_CEIL64, &ceil64, sizeof(ceil64));
#endif

	addattr_l(n, 2024, TCA_HTB_PARMS, &opt, sizeof(opt));
	addattr_l(n, 3024, TCA_HTB_RTAB, rtab, 1024);
	addattr_l(n, 4024, TCA_HTB_CTAB, ctab, 1024);
	tail->rta_len = (void *) NLMSG_TAIL(n) - (void *) tail;

	free_mem(rtab);
	free_mem(ctab);
	return 0;

error_exit:
	free_mem(rtab);
	free_mem(ctab);
	return -1;
}

