/*
 * Copyright 2014 Trend Micro Incorporated
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <net/sock.h>
#include <net/net_namespace.h>
#include <linux/workqueue.h>

#include "common.h"
#include "tc_common.h"

#define TC_WORKQUEUE 1

extern int qos_register_tc_cmd(int (*do_cmd)(char *));
extern void qos_unregister_tc_cmd(void);

static struct sock *rtnl_sock;
static struct sock *this_sock;

static struct net *init_net_ns;

#if TC_WORKQUEUE
typedef struct
{
	char cmd[256];
	struct work_struct ws;
} tc_work_t;

static struct workqueue_struct *tc_workqueue;

#endif

struct netlink_sock {
	/* struct sock has to be the first member of netlink_sock */
	struct sock		sk;
	u32			pid;
	u32			dst_pid;
	u32			dst_group;
	u32			flags;
	u32			subscriptions;
	u32			ngroups;
	unsigned long		*groups;
	unsigned long		state;
	wait_queue_head_t	wait;
	struct netlink_callback	*cb;
	struct mutex		*cb_mutex;
	struct mutex		cb_def_mutex;
	void			(*netlink_rcv)(struct sk_buff *skb);
	struct module		*module;
};

static inline struct netlink_sock *nlk_sk(struct sock *sk)
{
	return container_of(sk, struct netlink_sock, sk);
}

static int cmd2args(char **args, int args_size, char *cmd)
{
	int i;
	char *cur = cmd;
	const char *delim = " \t";

	for (i = 0;; i++)
	{
		if (i >= args_size)
		{
		    TC_ERR("command arguments overflow (%d out of %d)", i, args_size);
		    break;
		}
		args[i] = strsep(&cur, delim);
		if (args[i] == NULL)
		{
			break;
		}
	}

	return i;
}

static void this_rcv(struct sk_buff *skb)
{
	struct nlmsghdr *nlh;
	struct nlmsgerr *err;

	while (skb->len >= NLMSG_HDRLEN)
	{
		int msg_len;

		nlh = nlmsg_hdr(skb);

		if (nlh->nlmsg_len < NLMSG_HDRLEN || skb->len < nlh->nlmsg_len)
			break;

		if (nlh->nlmsg_type == NLMSG_ERROR)
		{
			err = nlmsg_data(nlh);
			if (err->error)
			{
				TC_ERR("Netlink return error: %d", err->error);
			}
			else
			{
				TC_DBG("Netlink return success: %d", err->error);
			}
		}

		msg_len = NLMSG_ALIGN(nlh->nlmsg_len);
		skb_pull(skb, msg_len);
	}
}

int do_unicast(struct sk_buff *skb)
{
	int ret;
	struct netlink_sock *rt_nlk;

	if (!init_net_ns)
	{
		TC_ERR("Cannot use init's network namespace!");
		return -1;
	}

	rtnl_sock = init_net_ns->rtnl;
	if (!rtnl_sock)
	{
		TC_ERR("Cannot get rtnetlink sock.");
		return -1;
	}

	ret = -1;
	rt_nlk = nlk_sk(rtnl_sock);
	if (rt_nlk && rt_nlk->netlink_rcv)
	{
		ret = skb->len;
		skb_set_owner_r(skb, this_sock);
		rt_nlk->netlink_rcv(skb);
	}
	else
	{
		TC_ERR("Cannot get rtnetlink receiver function.");
		return -1;
	}
	//kfree_skb(skb);	// Don't free skb here, free it where it was allocated
	// Don't do sock_put here, because no sock_hold.

	TC_DBG("Unicast %u bytes success.", ret);
	ret = 0;

	return ret;
}

#if TC_WORKQUEUE
void do_work(struct work_struct *ws)
{
	tc_work_t *work = container_of(ws, tc_work_t, ws);
	char *cmd;
	char *args[32] = { NULL };
	char **argv = args;
	char argc = 0;

	if (!work)
	{
		TC_ERR("Work not found!\n");
		return;
	}

	cmd = work->cmd;

	TC_DBG("do_work: %s", cmd);

	argc = cmd2args(args, sizeof(args) / sizeof(*args), cmd);

	if (argc <= 1)
	{
		TC_ERR("Arguments too short.");
		goto __free_exit;
	}

	if (strcmp(*argv, "qdisc") == 0)
	{
		do_qdisc(argc-1, argv+1);
		goto __free_exit;
	}

	if (strcmp(*argv, "class") == 0)
	{
		do_class(argc-1, argv+1);
		goto __free_exit;
	}

	if (strcmp(*argv, "filter") == 0)
	{
		do_filter(argc-1, argv+1);
		goto __free_exit;
	}

	TC_ERR("Unkown command type \"%s\".", *argv);

__free_exit:
	kfree(work);
	work = NULL;
	return;
}
#endif

int do_cmd(char *cmd)
{
#if TC_WORKQUEUE
	tc_work_t *tc_work;
	int cmd_len = strlen(cmd);

	if (!(tc_work = kmalloc(sizeof(tc_work_t), GFP_ATOMIC)))
	{
		MEM_ERR(sizeof(tc_work_t));
		return -1;
	}

	memcpy(tc_work->cmd, cmd, cmd_len);
	tc_work->cmd[cmd_len] = 0;

	TC_DBG("do_cmd: %s", tc_work->cmd);

	INIT_WORK(&tc_work->ws, do_work);

	queue_work(tc_workqueue, &tc_work->ws);

	return 0;
#else
	char *args[32] = { NULL };
	char **argv = args;
	char argc = 0;

	argc = cmd2args(args, sizeof(args) / sizeof(*args), cmd);

	if (argc <= 1)
	{
		TC_ERR("Arguments too short.");
		return -1;
	}

	if (strcmp(*argv, "qdisc") == 0)
	{

		return do_qdisc(argc-1, argv+1);
	}

	if (strcmp(*argv, "class") == 0)
	{
		return do_class(argc-1, argv+1);
	}

	if (strcmp(*argv, "filter") == 0)
	{
		return do_filter(argc-1, argv+1);
	}

	TC_ERR("Unkown command type \"%s\".", *argv);
	return -1;
#endif
}

int init_tc(struct module *module)
{
#if TC_WORKQUEUE
	tc_workqueue = create_singlethread_workqueue("tc_wq");

	if (!tc_workqueue)
	{
		TC_ERR("Cannot create workqueue for tc!");
		return -ENOENT;
	}
#endif

	init_net_ns = &(init_net);

	if (!init_net_ns)
	{
		TC_ERR("Cannot use init's network namespace!");
		return -ENOENT;
	}

	this_sock = netlink_kernel_create(init_net_ns, 21 /* Hardcoded */, 0, &this_rcv, NULL, module);

	if (!this_sock)
	{
		TC_ERR("Cannot create receiver sock.");
		return -EBUSY;
	}

	return 0;
}

void deinit_tc(void)
{
#if TC_WORKQUEUE
	if (tc_workqueue)
	{
		flush_workqueue(tc_workqueue);
		destroy_workqueue(tc_workqueue);
	}
#endif

	if (this_sock)
	{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
		netlink_kernel_release(this_sock);
#else
		sock_release(this_sock->sk_socket);
#endif
	}
}

int tc_init(void)
{
	int ret = 0;

	init_tc(THIS_MODULE);

	if ((ret = qos_register_tc_cmd(do_cmd)))
	{
		TC_ERR("Failed to register function to qos!\n");
	}

	return ret;
}

void tc_exit(void)
{
	deinit_tc();
	qos_unregister_tc_cmd();
}

MODULE_LICENSE("GPL");

module_init(tc_init);
module_exit(tc_exit);

