/*
 * Copyright 2014 Trend Micro Incorporated
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/
 */

#include <linux/kernel.h>
#include "tc_common.h"

int class_modify(int cmd, unsigned flags, int argc, char **argv)
{
	struct sk_buff *skb;
	struct nl_req *req;

	int req_len = sizeof(*req) + (sizeof(*req->buf) * 4096);
	int data_len;
	int ret;

	char dev[16];
	char kind[16];
	struct qdisc_util *q = NULL;

	/* Prepare buffer and initialize it */
	alloc_mem(req, req_len, GFP_KERNEL);

	memset(req, 0, req_len);
	memset(&dev, 0, sizeof(dev));
	memset(&kind, 0, sizeof(kind));

	req->n.nlmsg_len = NLMSG_LENGTH(sizeof(struct tcmsg));
	req->n.nlmsg_flags = NLM_F_REQUEST | NLM_F_ACK | flags;
	req->n.nlmsg_type = cmd;
	req->t.tcm_family = AF_UNSPEC;

	/* Parsing options */
	while (argc > 0)
	{
		if (strcmp(*argv, "dev") == 0)
		{
			NEXT_ARG();
			if (!dev[0])
			{
				strncpy(dev, *argv, sizeof(dev) - 1);
			}
		}
		else if (strcmp(*argv, "classid") == 0)
		{
			u32 handle;
			NEXT_ARG();
			if (get_handle(&handle, *argv))
			{
				TC_ERR("Invalid class ID \"%s\".", *argv);
				goto error_exit;
			}
			req->t.tcm_handle = handle;
		}
		else if (strcmp(*argv, "handle") == 0)
		{
			TC_ERR("Try \"classid\" instead of \"handle\".");
			goto error_exit;
		}
		else if (strcmp(*argv, "root") == 0)
		{
			req->t.tcm_parent = TC_H_ROOT;
		}
		else if (strcmp(*argv, "parent") == 0)
		{
			u32 handle;
			NEXT_ARG();
			if (get_handle(&handle, *argv))
			{
				TC_ERR("Invalid parent ID \"%s\".", *argv);
				goto error_exit;
			}
			req->t.tcm_parent = handle;
		}
		else
		{
			/* Get qdisc kind and qdisc utils */
			strncpy(kind, *argv, sizeof(kind) - 1);
			q = get_qdisc_util(kind);
			argc--;
			argv++;
			break;
		}

		argc--;
		argv++;
	}

	if (kind[0])
	{
		addattr_l(&req->n, req_len, TCA_KIND, kind, strlen(kind) + 1);
	}

	if (q)	/* Call qdisc util to parse class options */
	{
		if (q->parse_copt == NULL)
		{
			TC_ERR("Error: qdisc \"%s\" is classless.", kind);
			goto error_exit;
		}
		if (q->parse_copt(q, argc, argv, &req->n))
		{
			TC_ERR("Cannot parse qdisc \"%s\" class options.", kind);
			goto error_exit;
		}
	}
	else
	{
		if (argc)
		{
			TC_ERR("Garbage arguments \"%s ...\"", *argv);
			goto error_exit;
		}
	}

	if (dev[0])	/* Get ifindex */
	{
		int ifindex = get_ifindex_by_name(dev);
		if (ifindex <= 0)
		{
			goto error_exit;
		}
		req->t.tcm_ifindex = ifindex;
	}

	/* Prepare data for skb */
	data_len = NLMSG_ALIGN(req->n.nlmsg_len);
	skb = prepare_req_skb(data_len, req);
	free_mem(req);

	ret = -1;
	ret = do_unicast(skb);

	free_skb(skb);
	return ret;

error_exit:
	free_mem(req);
	return -1;
}

int do_class(int argc, char **argv)
{
	if (argc <= 1)
	{
		TC_ERR("Arguments too short.");
		return -1;
	}

	if (matches(*argv, "add") == 0)
	{
		return class_modify(RTM_NEWTCLASS, NLM_F_EXCL|NLM_F_CREATE, argc-1, argv+1);
	}
	if (matches(*argv, "change") == 0)
	{
		return class_modify(RTM_NEWTCLASS, 0, argc-1, argv+1);
	}
	if (matches(*argv, "replace") == 0)
	{
		return class_modify(RTM_NEWTCLASS, NLM_F_CREATE, argc-1, argv+1);
	}
	if (matches(*argv, "delete") == 0)
	{
		return class_modify(RTM_DELTCLASS, 0, argc-1, argv+1);
	}

	TC_ERR("Unkown command \"%s\".", *argv);
	return -1;
}

