/*
 * Copyright 2014 Trend Micro Incorporated
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/
 */

#include <linux/kernel.h>
#include <linux/version.h>
#include "tc_common.h"

/* Ref: linux/if_ether.h */
#define __PF(f,n) { ETH_P_##f, #n },
static const struct {
	int id;
	const char *name;
} llproto_names[] = {
__PF(LOOP,loop)
__PF(PUP,pup)
__PF(PUPAT,pupat)
__PF(IP,ip)
__PF(X25,x25)
__PF(ARP,arp)
__PF(BPQ,bpq)
__PF(IEEEPUP,ieeepup)
__PF(IEEEPUPAT,ieeepupat)
__PF(DEC,dec)
__PF(DNA_DL,dna_dl)
__PF(DNA_RC,dna_rc)
__PF(DNA_RT,dna_rt)
__PF(LAT,lat)
__PF(DIAG,diag)
__PF(CUST,cust)
__PF(SCA,sca)
__PF(RARP,rarp)
__PF(ATALK,atalk)
__PF(AARP,aarp)
__PF(IPX,ipx)
__PF(IPV6,ipv6)
__PF(PPP_DISC,ppp_disc)
__PF(PPP_SES,ppp_ses)
__PF(ATMMPOA,atmmpoa)
__PF(ATMFATE,atmfate)
__PF(802_3,802_3)
__PF(AX25,ax25)
__PF(ALL,all)
__PF(802_2,802_2)
__PF(SNAP,snap)
__PF(DDCMP,ddcmp)
__PF(WAN_PPP,wan_ppp)
__PF(PPP_MP,ppp_mp)
__PF(LOCALTALK,localtalk)
__PF(CAN,can)
__PF(PPPTALK,ppptalk)
__PF(TR_802_2,tr_802_2)
__PF(MOBITEX,mobitex)
__PF(CONTROL,control)
__PF(IRDA,irda)
__PF(ECONET,econet)
__PF(TIPC,tipc)
__PF(AOE,aoe)
__PF(8021Q,802.1Q)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0))
__PF(8021AD,802.1ad)
#endif

{ 0x8100, "802.1Q" },
{ 0x88cc, "LLDP" },
{ ETH_P_IP, "ipv4" },
};
#undef __PF

const char *ll_proto_n2a(unsigned short id, char *buf, int len)
{
	int i;
	id = ntohs(id);

	for (i = 0; i < sizeof(llproto_names) / sizeof(llproto_names[0]); i++)
	{
		if (llproto_names[i].id == id)
		{
			return llproto_names[i].name;
		}
	}
	snprintf(buf, len, "[%d]", id);
	return buf;
}

int ll_proto_a2n(unsigned short *id, const char *buf)
{
	int i;
	for (i = 0; i < sizeof(llproto_names) / sizeof(llproto_names[0]); i++)
	{
		if (strcasecmp(llproto_names[i].name, buf) == 0)
		{
			*id = htons(llproto_names[i].id);
			return 0;
		}
	}
	if (get_u16(id, buf, 0))
	{
		return -1;
	}
	*id = htons(*id);
	return 0;
}

static int filter_modify(int cmd, unsigned flags, int argc, char **argv)
{
	struct sk_buff *skb;
 	struct nl_req *req;

	int req_len = sizeof(*req) + (sizeof(*req->buf) * MAX_MSG);
	int data_len;
	int ret;

	char dev[16];
	char kind[16];
	struct filter_util *f = NULL;

	u32 prio = 0;
	u32 protocol = 0;
	char *fhandle = NULL;

	/* Prepare buffer and initialize it */
	alloc_mem(req, req_len, GFP_KERNEL);

	memset(req, 0, req_len);
	memset(&dev, 0, sizeof(dev));
	memset(&kind, 0, sizeof(kind));

	req->n.nlmsg_len = NLMSG_LENGTH(sizeof(struct tcmsg));
	req->n.nlmsg_flags = NLM_F_REQUEST | NLM_F_ACK | flags;
	req->n.nlmsg_type = cmd;
	req->t.tcm_family = AF_UNSPEC;

	if (cmd == RTM_NEWTFILTER && flags & NLM_F_CREATE)
	{
		protocol = htons(ETH_P_ALL);
	}

	/* Parsing options */
	while (argc > 0)
	{
		if (strcmp(*argv, "dev") == 0)
		{
			NEXT_ARG();
			strncpy(dev, *argv, sizeof(dev) - 1);
		}
		else if (strcmp(*argv, "root") == 0)
		{
			req->t.tcm_parent = TC_H_ROOT;
		}
		else if (strcmp(*argv, "parent") == 0)
		{
			u32 handle;
			NEXT_ARG();
			if (get_handle(&handle, *argv))
			{
				TC_ERR("Invalid parent ID \"%s\".", *argv);
				goto error_exit;
			}
			req->t.tcm_parent = handle;
		}
		else if (strcmp(*argv, "handle") == 0)
		{
			NEXT_ARG();
			fhandle = *argv;
		}
		else if (matches(*argv, "preference") == 0 ||
			   matches(*argv, "priority") == 0)
		{
			NEXT_ARG();
			if (get_u32(&prio, *argv, 0) || prio > 0xFFFF)
			{
				TC_ERR("Invalid priority value \"%s\".", *argv);
				goto error_exit;
			}
		}
		else if (matches(*argv, "protocol") == 0)
		{
			u16 id;
			NEXT_ARG();
			if (ll_proto_a2n(&id, *argv))
			{
				TC_ERR("invalid protocol \"%s\".", *argv);
				goto error_exit;
			}
			protocol = id;
		}
		else
		{
			/* Get filter kind and filter utils */
			strncpy(kind, *argv, sizeof(kind) - 1);
			f = get_filter_util(kind);
			argc--;
			argv++;
			break;
		}

		argc--;
		argv++;
	}

	req->t.tcm_info = TC_H_MAKE(prio << 16, protocol);

	if (kind[0])
	{
		addattr_l(&req->n, req_len, TCA_KIND, kind, strlen(kind) + 1);
	}

	if (f)	/* Call filter util to parse filter options */
	{
		if (f->parse_fopt(f, fhandle, argc, argv, &req->n))
		{
			TC_ERR("Cannot parse %s filter options.", kind);
			goto error_exit;
		}
	}
	else
	{
		if (fhandle)
		{
			TC_ERR("Must specify filter type when using \"handle\".");
			goto error_exit;
		}
		if (argc)
		{
			TC_ERR("Garbage arguments \"%s ...\"", *argv);
			goto error_exit;
		}
	}

	if (dev[0])	/* Get ifindex */
	{
		int ifindex = get_ifindex_by_name(dev);
		if (ifindex <= 0)
		{
			goto error_exit;
		}
		req->t.tcm_ifindex = ifindex;
	}

	/* Prepare data for skb */
	data_len = NLMSG_ALIGN(req->n.nlmsg_len);
	skb = prepare_req_skb(data_len, req);
	free_mem(req);

	ret = -1;
	ret = do_unicast(skb);

	free_skb(skb);
	return ret;

error_exit:
	free_mem(req);
	return -1;
}

int do_filter(int argc, char **argv)
{
	if (argc <= 1)
	{
		TC_ERR("Arguments too short.");
		return -1;
	}

	if (matches(*argv, "add") == 0)
	{
		return filter_modify(RTM_NEWTFILTER, NLM_F_EXCL | NLM_F_CREATE, argc-1, argv+1);
	}
	if (matches(*argv, "change") == 0)
	{
		return filter_modify(RTM_NEWTFILTER, 0, argc-1, argv+1);
	}
	if (matches(*argv, "replace") == 0)
	{
		return filter_modify(RTM_NEWTFILTER, NLM_F_CREATE, argc-1, argv+1);
	}
	if (matches(*argv, "delete") == 0)
	{
		return filter_modify(RTM_DELTFILTER, 0, argc-1, argv+1);
	}

	TC_ERR("Unkown command \"%s\".", *argv);
	return -1;
}
