/*
 * Copyright 2014 Trend Micro Incorporated
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/
 */

#include <linux/kernel.h>
#include <linux/atm.h>
#include <linux/capability.h>
#include <linux/time.h>
#include "tc_common.h"

#define TIME_UNITS_PER_SEC	1000000
#define tick_in_usec (u32)NSEC_PER_USEC / (u32)PSCHED_TICKS2NS(1)

static struct qdisc_util q_util[] = {
	{"htb", htb_parse_opt, htb_parse_class_opt}
#ifdef SFQ_OPT_SUPPORT
	, {"sfq", sfq_parse_opt, NULL}
#endif
#ifdef FIFO_OPT_SUPPORT
	, {"bfifo", fifo_parse_opt, NULL}
	, {"pfifo", fifo_parse_opt, NULL}
#endif
#ifdef PRIO_OPT_SUPPORT
	, {"prio", prio_parse_opt, NULL}
#endif
};

static struct filter_util f_util[] = {
	{"u32", u32_parse_filter_opt}
};

int matches(const char *cmd, const char *pattern)
{
	int len = strlen(cmd);
	if (len > strlen(pattern))
	{
		return -1;
	}
	return memcmp(pattern, cmd, len);
}

int get_hz(void)
{
	struct timespec ts;
	hrtimer_get_res(CLOCK_MONOTONIC, &ts);
#if 0
	TC_DBG("%u %u %u %u",
		(u32)NSEC_PER_USEC,
		(u32)PSCHED_TICKS2NS(1),
		1000000,
		(u32)NSEC_PER_SEC / (u32)ktime_to_ns(timespec_to_ktime(ts)));
#endif
	TC_DBG("tick_in_usec = %u / %u", (u32)NSEC_PER_USEC, (u32)PSCHED_TICKS2NS(1));
	return (u32)NSEC_PER_SEC / (u32)ktime_to_ns(timespec_to_ktime(ts));
}

int get_unsigned(unsigned *val, const char *arg, int base)
{
	unsigned long res;
	char *ptr;

	if (!arg || !*arg)
		return -1;

	res = strtoul(arg, &ptr, base);

	/* empty string or trailing non-digits */
	if (!ptr || ptr == arg || *ptr)
		return -1;

	/* overflow */
	if (res == ULONG_MAX)
		return -1;

	/* out side range of unsigned */
	if (res > UINT_MAX)
		return -1;

	*val = res;
	return 0;
}

int get_u32(u32 *val, const char *arg, int base)
{
	unsigned long res;
	char *ptr;

	if (!arg || !*arg)
		return -1;

	res = strtoul(arg, &ptr, base);

	/* empty string or trailing non-digits */
	if (!ptr || ptr == arg || *ptr)
		return -1;

	/* overflow */
	if (res == ULONG_MAX)
		return -1;

	/* in case UL > 32 bits */
	if (res > 0xFFFFFFFFUL)
		return -1;

	*val = res;
	return 0;
}

int get_u16(u16 *val, const char *arg, int base)
{
	unsigned long res;
	char *ptr;

	if (!arg || !*arg)
		return -1;

	res = strtoul(arg, &ptr, base);

	/* empty string or trailing non-digits */
	if (!ptr || ptr == arg || *ptr)
		return -1;

	/* overflow */
	if (res == ULONG_MAX)
		return -1;

	if (res > 0xFFFFUL)
		return -1;

	*val = res;
	return 0;
}

int get_integer(int *val, const char *arg, int base)
{
	long res;
	char *ptr;

	if (!arg || !*arg)
		return -1;

	res = strtol(arg, &ptr, base);

	/* If there were no digits at all, strtol()  stores
         * the original value of nptr in *endptr (and returns 0).
	 * In particular, if *nptr is not '\0' but **endptr is '\0' on return,
	 * the entire string is valid.
	 */
	if (!ptr || ptr == arg || *ptr)
		return -1;

	/* If an underflow occurs, strtol() returns LONG_MIN.
	 * If an overflow occurs,  strtol() returns LONG_MAX.
	 * In both cases, errno is set to ERANGE.
	 */
	if ((res == LONG_MAX || res == LONG_MIN))
		return -1;

	/* Outside range of int */
	if (res < INT_MIN || res > INT_MAX)
		return -1;

	*val = res;
	return 0;
}

static const struct rate_suffix {
	const char *name;
	unsigned long long scale;
} suffixes[] = {
	{ "bit",	1. },
	{ "Kibit",	1024. },
	{ "kbit",	1000. },
	{ "mibit",	1024.*1024. },
	{ "mbit",	1000000. },
	{ "gibit",	1024.*1024.*1024. },
	{ "gbit",	1000000000. },
	{ "tibit",	1024.*1024.*1024.*1024. },
	{ "tbit",	1000000000000. },
	{ "Bps",	8. },
	{ "KiBps",	8.*1024. },
	{ "KBps",	8000. },
	{ "MiBps",	8.*1024*1024. },
	{ "MBps",	8000000. },
	{ "GiBps",	8.*1024.*1024.*1024. },
	{ "GBps",	8000000000. },
	{ "TiBps",	8.*1024.*1024.*1024.*1024. },
	{ "TBps",	8000000000000. },
	{ NULL }
};

int get_rate(unsigned *rate, const char *str)
{
	char *p;
	unsigned long long bps = strtoull(str, &p, 0);
	const struct rate_suffix *s;

	if (p == str)
		return -1;

	for (s = suffixes; s->name; ++s)
	{
		if (strcasecmp(s->name, p) == 0)
		{
			bps *= s->scale;
			p += strlen(p);
			break;
		}
	}

	if (*p)
		return -1; /* unknown suffix */

	bps /= 8; /* -> bytes per second */
	*rate = bps;

	if (*rate >= UINT_MAX)
		return -1;
	/* detect if an overflow happened */
	//if (*rate != floor(bps))
	//	return -1;
	return 0;
}

int get_rate64(__u64 *rate, const char *str)
{
	char *p;
	unsigned long long bps = strtoull(str, &p, 0);
	const struct rate_suffix *s;

	if (p == str)
		return -1;

	for (s = suffixes; s->name; ++s)
	{
		if (strcasecmp(s->name, p) == 0)
		{
			bps *= s->scale;
			p += strlen(p);
			break;
		}
	}

	if (*p)
		return -1; /* unknown suffix */

	bps /= 8; /* -> bytes per second */
	*rate = bps;
	return 0;
}


int get_size(unsigned *size, const char *str)
{
	unsigned long sz;
	char *p;

	sz = strtoul(str, &p, 0);
	if (p == str)
		return -1;

	if (*p) {
		if (strcasecmp(p, "kb") == 0 || strcasecmp(p, "k")==0)
			sz *= 1024;
		else if (strcasecmp(p, "gb") == 0 || strcasecmp(p, "g")==0)
			sz *= 1024*1024*1024;
		else if (strcasecmp(p, "gbit") == 0)
			sz *= 1024*1024*1024/8;
		else if (strcasecmp(p, "mb") == 0 || strcasecmp(p, "m")==0)
			sz *= 1024*1024;
		else if (strcasecmp(p, "mbit") == 0)
			sz *= 1024*1024/8;
		else if (strcasecmp(p, "kbit") == 0)
			sz *= 1024/8;
		else if (strcasecmp(p, "b") != 0)
			return -1;
	}

	if (sz >= UINT_MAX)
		return -1;

	*size = sz;
	return 0;
}

int get_size_and_cell(unsigned *size, int *cell_log, char *str)
{
	char * slash = strchr(str, '/');

	if (slash)
		*slash = 0;

	if (get_size(size, str))
		return -1;

	if (slash) {
		int cell;
		int i;

		if (get_integer(&cell, slash+1, 0))
			return -1;
		*slash = '/';

		for (i=0; i<32; i++) {
			if ((1<<i) == cell) {
				*cell_log = i;
				return 0;
			}
		}
		return -1;
	}
	return 0;
}

int addattr_l(struct nlmsghdr *n, int maxlen, int type, const void *data,
	      int alen)
{
	int len = RTA_LENGTH(alen);
	struct rtattr *rta;

	TC_DBG("add attr len = %d", RTA_ALIGN(len));

	if (NLMSG_ALIGN(n->nlmsg_len) + RTA_ALIGN(len) > maxlen) {
		TC_ERR("addattr_l ERROR: message exceeded bound of %d", maxlen);
		return -1;
	}
	rta = NLMSG_TAIL(n);
	rta->rta_type = type;
	rta->rta_len = len;
	memcpy(RTA_DATA(rta), data, alen);
	n->nlmsg_len = NLMSG_ALIGN(n->nlmsg_len) + RTA_ALIGN(len);
	return 0;
}

struct rtattr *addattr_nest(struct nlmsghdr *n, int maxlen, int type)
{
	struct rtattr *nest = NLMSG_TAIL(n);

	addattr_l(n, maxlen, type, NULL, 0);
	return nest;
}

int addattr_nest_end(struct nlmsghdr *n, struct rtattr *nest)
{
	nest->rta_len = (void *)NLMSG_TAIL(n) - (void *)nest;
	return n->nlmsg_len;
}

struct rtattr *addattr_nest_compat(struct nlmsghdr *n, int maxlen, int type,
				   const void *data, int len)
{
	struct rtattr *start = NLMSG_TAIL(n);

	addattr_l(n, maxlen, type, data, len);
	addattr_nest(n, maxlen, type);
	return start;
}

int addattr_nest_compat_end(struct nlmsghdr *n, struct rtattr *start)
{
	struct rtattr *nest = (void *)start + NLMSG_ALIGN(start->rta_len);

	start->rta_len = (void *)NLMSG_TAIL(n) - (void *)start;
	addattr_nest_end(n, nest);
	return n->nlmsg_len;
}

int get_ifindex_by_name(const char *dev)
{
	struct net_device *net_dev;

	net_dev = __dev_get_by_name(&init_net, dev);
	if (net_dev)
	{
		return net_dev->ifindex;
	}
	else
	{
		TC_ERR("Cannot find device \"%s\".", dev);
		return -1;
	}
}

struct qdisc_util *get_qdisc_util(const char *str)
{
	int qu_size = sizeof(q_util) / sizeof(*q_util);
	struct qdisc_util *q = NULL;
	int i = 0;

	for (i = 0; i < qu_size; i++)
	{
		if (strcmp(str, q_util[i].id) == 0)
		{
			q = &q_util[i];
		}
	}

	return q;
}

struct filter_util *get_filter_util(const char *str)
{
	int fu_size = sizeof(f_util) / sizeof(*f_util);
	struct filter_util *f = NULL;
	int i = 0;

	for (i = 0; i < fu_size; i++)
	{
		if (strcmp(str, f_util[i].id) == 0)
		{
			f = &f_util[i];
		}
	}

	return f;
}

int get_handle(u32 *h, const char *str)
{
	u32 maj, min;
	char *p;

	if (strcmp(str, "root") == 0)
	{
		maj = TC_H_ROOT;
		goto ok;
	}
	if (strcmp(str, "none") == 0)
	{
		maj = TC_H_UNSPEC;
		goto ok;
	}

	maj = strtoul(str, &p, 16);
	if (p == str)
	{
		maj = 0;
		if (*p != ':')
			goto nok;
	}

	if (*p == ':')
	{
		if (maj >= (1<<16))
		{
			goto nok;
		}
		maj <<= 16;
		str = p + 1;
		min = strtoul(str, &p, 16);
		if (*p != 0 || min >= (1 << 16))
		{
			goto nok;
		}
		maj |= min;
	}
	else if (*p != 0)
	{
		goto nok;
	}

ok:
	*h = maj;
	return 0;

nok:
	return -1;
}

#if 0
/*
   Plog = log(prob/(qmax - qmin))
 */
int tc_red_eval_P(unsigned qmin, unsigned qmax, double prob)
{
	int i = qmax - qmin;

	if (i <= 0)
		return -1;

	prob /= i;

	for (i=0; i<32; i++) {
		if (prob > 1.0)
			break;
		prob *= 2;
	}
	if (i>=32)
		return -1;
	return i;
}

/*
   burst + 1 - qmin/avpkt < (1-(1-W)^burst)/W
 */

int tc_red_eval_ewma(unsigned qmin, unsigned burst, unsigned avpkt)
{
	int wlog = 1;
	double W = 0.5;
	double a = (double)burst + 1 - (double)qmin/avpkt;

	if (a < 1.0) {
		fprintf(stderr, "tc_red_eval_ewma() burst %u is too small ?"
				" Try burst %u\n", burst, 1 + qmin/avpkt);
		return -1;
	}
	for (wlog=1; wlog<32; wlog++, W /= 2) {
		if (a <= (1 - pow(1-W, burst))/W)
			return wlog;
	}
	return -1;
}
#endif

unsigned tc_core_time2tick(unsigned time)
{
	return time * tick_in_usec;
}

unsigned tc_calc_xmittime(u32 rate, unsigned size)
{
#define TIME_UNIT_ROOT 1000
#define UMAX_DIV_BY_TIME_UNIT 4294
#define UMAX_DIV_BY_TIME_UNIT_ROOT 4294967
	unsigned time;

	if (size > UMAX_DIV_BY_TIME_UNIT_ROOT)
	{
		time = TIME_UNITS_PER_SEC * (size / rate);
	}
	else if (size > UMAX_DIV_BY_TIME_UNIT)
	{
		time = TIME_UNIT_ROOT * ((TIME_UNIT_ROOT * size) / rate);
	}
	else
	{
		time = (TIME_UNITS_PER_SEC * size) / rate;
	}

	return tc_core_time2tick(time);
}

static unsigned tc_align_to_atm(unsigned size)
{
	int linksize, cells;
	cells = size / ATM_CELL_PAYLOAD;
	if ((size % ATM_CELL_PAYLOAD) > 0)
		cells++;

	linksize = cells * ATM_CELL_SIZE; /* Use full cell size to add ATM tax */
	return linksize;
}

static unsigned tc_adjust_size(unsigned sz, unsigned mpu, enum link_layer linklayer)
{
	if (sz < mpu)
		sz = mpu;

	switch (linklayer) {
	case LINKLAYER_ATM:
		return tc_align_to_atm(sz);
	case LINKLAYER_ETHERNET:
	default:
		// No size adjustments on Ethernet
		return sz;
	}
}

int tc_calc_rtable(struct __tc_ratespec *r, u32 *rtab,
		   int cell_log, unsigned mtu,
		   enum link_layer linklayer)
{
	int i;
	unsigned sz;
	unsigned bps = r->rate;
	unsigned mpu = r->mpu;

	if (mtu == 0)
		mtu = 2047;

	if (cell_log < 0) {
		cell_log = 0;
		while ((mtu >> cell_log) > 255)
			cell_log++;
	}

	for (i=0; i<256; i++) {
		sz = tc_adjust_size((i + 1) << cell_log, mpu, linklayer);
		rtab[i] = tc_calc_xmittime(bps, sz);
	}

	r->cell_align=-1; // Due to the sz calc
	r->cell_log=cell_log;
	r->linklayer = (linklayer & TC_LINKLAYER_MASK);
	return cell_log;
}

struct sk_buff *prepare_req_skb(int len, void *req)
{
	struct sk_buff *skb;
	void *data;

	skb = alloc_skb(len, GFP_KERNEL);
	if (!skb)
	{
		TC_ERR("Cannot allocate skb.");
		return NULL;
	}

	NETLINK_CB(skb).pid = 0; /* from kernel */
	NETLINK_CB(skb).dst_group = 0; /* unicast */
	cap_raise(NETLINK_CB(skb).eff_cap, CAP_NET_ADMIN);

	data = skb_put(skb, len);
	//TC_DBG("skb->len = %d", skb->len);
	memcpy(data, req, len);

	return skb;
}

